# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to check fdog data which are present in
#  searchTaxa_dir, coreTaxa_dir and annotation_dir
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import argparse
from os import listdir
from os.path import isfile, join
from pathlib import Path
import subprocess
import shutil
from Bio import SeqIO
from ete3 import NCBITaxa
import re
from datetime import datetime
import multiprocessing as mp
from tqdm import tqdm
from pkg_resources import get_distribution
from Bio.Blast.Applications import NcbiblastpCommandline


import fdog.libs.zzz as general_fn
import fdog.libs.blast as blast_fn
import fdog.libs.fasta as fasta_fn


def checkOptConflict(concat, replace, delete):
    if concat:
        if not (delete or replace):
            sys.exit('*** ERROR: for rewrite sequences, you need to set either "--delete" or "--replace"!')
    if delete:
        if replace:
            sys.exit('*** ERROR: only one option can be choose between "--replace" and "--delete"')
    if replace:
        if delete:
            sys.exit('*** ERROR: only one option can be choose between "--replace" and "--delete"')


def check_valid_fasta(file):
    """ Check if valid fasta file
    Check if:
    (1) Input file is a fasta file
    (2) If headers are longer than 30 characters
    (3) If headers and sequences contain any space/tab
    (4) If sequences are written in a single line
    """
    spaceChr = (' ', '\t')
    with open(file, 'r') as f:
        f_bkp = f
        # check if input file is a FASTA file
        fasta = SeqIO.parse(f, 'fasta')
        if not any(fasta):
            return({'notFasta': 1})
        else:
            # check for long header
            inSeq = SeqIO.to_dict((SeqIO.parse(open(file), 'fasta')))
            for id in inSeq:
                if len(id) > 30:
                    return({'longHeader': id})
        # check space or tab
        if any(s in f.read() for s in spaceChr):
            return('space')
    # check single line
    nHeader = general_fn.count_line(file, '>', True)
    nSeq = general_fn.count_line(file, '>', False)
    if not nHeader == nSeq:
        return({'multiLine': 1})
    return({'ok': 1})


def check_valid_folder_name(folder):
    """ Check if folder name contains any special characters """
    invalidChr = (' ','|','\t','\'','"','`','´','^','!','$','%','&')
    if any(e in folder for e in invalidChr):
        sys.exit('*** ERROR: Invalid character found in %s' % folder)


def check_valid_seqs(fa_file):
    """ Check if any sequence contains space/tab or special characters """
    spaceChr = (' ', '\t')
    faSeq = SeqIO.parse(open(fa_file),'fasta')
    for fa in faSeq:
        id, seq = fa.description, str(fa.seq)
        c = ''
        if any(e in id for e in spaceChr):
            sys.exit('*** ERROR: Invalid character found in \">%s\" in %s' % (id, fa_file))
        if any(c for c in seq if not c.isalpha()):
            print('*** ERROR: Invalid character "%s" found in the sequence of gene \"%s\" in %s' % (c, id, fa_file))
            sys.exit('You can use "--replace" or "--delete" to solve this issue!')


def rewrite_seqs(fa_file, replace, delete):
    """ Rewrite fasta sequence by replacing or deleting special characters """
    spaceChr = (' ', '\t')
    faSeq = SeqIO.parse(open(fa_file),'fasta')
    with open(fa_file + '.mod', 'w') as tmpOut:
        for fa in faSeq:
            id, seq = fa.description, str(fa.seq)
            if replace:
                seq = re.sub('[^a-zA-Z]', 'X', seq)
            if delete:
                seq = re.sub('[^a-zA-Z]', '', seq)
            tmpOut.write('>%s\n%s\n' % (id, seq))
    os.replace(fa_file + '.mod', fa_file)


def write_faChecked(fa_file):
    """ Add fa.checked file in searchTaxa_dir """
    with open(fa_file+'.checked', 'w') as f:
        f.write(str(datetime.now()))


def check_fasta(args):
    """ Check fasta file in searchTaxa_dir and coreTaxa_dir """
    # taxon_list = []
    # for fd in general_fn.read_dir(checkDir):
    #     taxon = fd
    #     check_valid_folder_name('%s/%s' % (checkDir, taxon))
    #     for file in listdir('%s/%s' % (checkDir, taxon)):
    #         if file.endswith('.fa'):
    #             fa_file = '%s/%s/%s' % (checkDir, taxon, file)
    #             if os.path.islink(fa_file):
    #                 fa_file = os.path.realpath(fa_file)
    #             general_fn.check_file_exist(fa_file)
    #             checkfa_file = check_valid_fasta(fa_file)
    #             taxon_list.append(taxon)
    #             if not os.path.exists('%s.checked' % fa_file):
    #                 print(taxon)
    #                 if list(checkfa_file.keys())[0] == 'notFasta':
    #                     sys.exit('*** ERROR: %s does not look like a fasta file!' % fa_file)
    #                 elif list(checkfa_file.keys())[0] == 'longHeader':
    #                     sys.exit('*** ERROR: %s contains long headers! E.g. %s' % (fa_file, list(checkfa_file.values())[0]))
    #                 elif list(checkfa_file.keys())[0] == 'space':
    #                     sys.exit('*** ERROR: %s contains spaces/tabs!' % fa_file)
    #                 elif list(checkfa_file.keys())[0] == 'multiLine':
    #                     if not concat:
    #                         print('*** ERROR: %s contains multiple-line sequences!' % fa_file)
    #                         sys.exit('Please use "--concat" with "--replace" or "--delete" to join them into single lines')
    #                     else:
    #                         rewrite_seqs(fa_file, replace, delete)
    #                 elif list(checkfa_file.keys())[0] == 'ok':
    #                     if not (delete or replace):
    #                         check_valid_seqs(fa_file)
    #                     else:
    #                         rewrite_seqs(fa_file, replace, delete)
    #                 write_faChecked(fa_file)
    #             if not os.path.exists('%s.fai' % fa_file):
    #                 fasta_fn.read_fasta(fa_file)
    # return(taxon_list)
    (taxon, file, checkDir, replace, delete, concat) = args
    fa_file = '%s/%s/%s' % (checkDir, taxon, file)
    if os.path.islink(fa_file):
        fa_file = os.path.realpath(fa_file)
    general_fn.check_file_exist(fa_file)
    checkfa_file = check_valid_fasta(fa_file)
    # taxon_list.append(taxon)
    if not os.path.exists('%s.checked' % fa_file):
        print(taxon)
        if list(checkfa_file.keys())[0] == 'notFasta':
            sys.exit('*** ERROR: %s does not look like a fasta file!' % fa_file)
        elif list(checkfa_file.keys())[0] == 'longHeader':
            sys.exit('*** ERROR: %s contains long headers! E.g. %s' % (fa_file, list(checkfa_file.values())[0]))
        elif list(checkfa_file.keys())[0] == 'space':
            sys.exit('*** ERROR: %s contains spaces/tabs!' % fa_file)
        elif list(checkfa_file.keys())[0] == 'multiLine':
            if not concat:
                print('*** ERROR: %s contains multiple-line sequences!' % fa_file)
                sys.exit('Please use "--concat" with "--replace" or "--delete" to join them into single lines')
            else:
                rewrite_seqs(fa_file, replace, delete)
        elif list(checkfa_file.keys())[0] == 'ok':
            if not (delete or replace):
                check_valid_seqs(fa_file)
            else:
                rewrite_seqs(fa_file, replace, delete)
        write_faChecked(fa_file)
    if not os.path.exists('%s.fai' % fa_file):
        fasta_fn.read_fasta(fa_file)
    return(taxon)


def run_check_fasta(checkDir, replace, delete, concat):
    """ Run check_fasta fn """
    jobs = []
    for taxon in general_fn.read_dir(checkDir):
        check_valid_folder_name('%s/%s' % (checkDir, taxon))
        for file in listdir('%s/%s' % (checkDir, taxon)):
            if file.endswith('.fa'):
                jobs.append([taxon, file, checkDir, replace, delete, concat])
    cpus = mp.cpu_count()-1
    pool = mp.Pool(cpus)
    taxon_list = []
    for _ in tqdm(pool.imap_unordered(check_fasta, jobs), total=len(jobs)):
        taxon_list.append(_)
    return(taxon_list)


def check_blastdb(args):
    """ Check for outdated blastdb """
    (query, taxon, coreTaxa_dir) = args
    blast_db = '%s/%s/%s' % (coreTaxa_dir, taxon, taxon)
    try:
        blastp_cline = NcbiblastpCommandline(query = query, db = blast_db)
        stdout, stderr = blastp_cline()
    except:
        return([query, blast_db])
    if not os.path.exists('%s/%s/%s.fa.fai' % (coreTaxa_dir, taxon, taxon)):
        fai_in_genome = "../../searchTaxa_dir/%s/%s.fa.fai" % (taxon, taxon)
        fai_in_blast = "%s/%s/%s.fa.fai" % (coreTaxa_dir, taxon, taxon)
        os.symlink(fai_in_genome, fai_in_blast)


def run_check_blastdb(coreTaxa_dir, fdogPath):
    """ Run check_blastdb fn """
    query = '%s/data/infile.fa' % fdogPath
    jobs = []
    for fd in general_fn.read_dir(coreTaxa_dir):
        jobs.append([query, fd, coreTaxa_dir])
    cpus = mp.cpu_count()-1
    pool = mp.Pool(cpus)
    out = []
    for _ in tqdm(pool.imap_unordered(check_blastdb, jobs), total=len(jobs)):
        out.append(_)
    return([1])


def create_blastdb(args):
    """ Redo (or update) blastdb """
    (taxon, coreTaxa_dir, searchTaxa_dir, outPath) = args
    fa_file = '%s/%s/%s.fa' % (coreTaxa_dir, taxon, taxon)
    if os.path.islink(fa_file):
        fa_file = os.path.realpath(fa_file)
    if not os.path.exists(fa_file):
        fa_file = '%s/%s/%s.fa' % (searchTaxa_dir, taxon, taxon)
    if os.path.exists(fa_file):
        ### remove old files
        blast_path = '%s/%s' % (coreTaxa_dir, taxon)
        shutil.rmtree(blast_path)
        ### Redo blastdb
        Path(blast_path).mkdir(parents = True, exist_ok = True)
        blast_fn.make_blastdb([taxon, fa_file, outPath, True])
        ### make symlink to fasta files
        fa_in_genome = "../../searchTaxa_dir/%s/%s.fa" % (taxon, taxon)
        fai_in_genome = "../../searchTaxa_dir/%s/%s.fa.fai" % (taxon, taxon)
        fa_in_blast = "%s/%s.fa" % (blast_path, taxon)
        fai_in_blast = "%s/%s.fa.fai" % (blast_path, taxon)
        if not os.path.exists(fa_in_blast):
            os.symlink(fa_in_genome, fa_in_blast)
        if not os.path.exists(fai_in_blast):
            os.symlink(fai_in_genome, fai_in_blast)
        return None
    else:
        return(taxon)


def run_create_blastdb(coreTaxa_dir, searchTaxa_dir):
    """ Run create_blastdb fn """
    outPath = '/'.join(coreTaxa_dir.split('/')[0:-1])
    jobs = []
    for fd in general_fn.read_dir(coreTaxa_dir):
        jobs.append([fd, coreTaxa_dir, searchTaxa_dir, outPath])
    cpus = mp.cpu_count()-1
    pool = mp.Pool(cpus)
    out = []
    for _ in tqdm(pool.imap_unordered(create_blastdb, jobs), total=len(jobs)):
        out.append(_)
    return([i for i in out if i is not None])


def check_missing_json(annotation_dir, taxon_list):
    """ Check missing annotation for any taxa in coreTaxa_dir and searchTaxa_dir """
    all_anno = [f for f in listdir(annotation_dir) if isfile(join(annotation_dir, f))]
    taxa_anno = [s + '.json' for s in taxon_list]
    s = set(all_anno)
    missing_anno = [x for x in taxa_anno if x not in s]
    return(missing_anno)


def check_complete_anno(args):
    """ Check if an annotation is complete
    I.e. if it contains annotation for all proteins of a species
    """
    (gf,jf, annotation_dir) = args
    cmd = 'fas.checkAnno -s %s -a %s -o %s' % (gf, jf, annotation_dir)
    try:
        subprocess.call([cmd], shell = True, stdout=subprocess.DEVNULL)
    except subprocess.CalledProcessError as e:
        print('*** ERROR: Problem while checking annotation file using fas.checkAnno!')
        print(e.output.decode(sys.stdout.encoding))
        sys.exit()


def run_check_complete_anno(annotation_dir, searchTaxa_dir):
    """ Run check_complete_anno fn """
    all_anno = [f for f in listdir(annotation_dir) if isfile(join(annotation_dir, f))]
    jobs = []
    for f in all_anno:
        tax = f.replace('.json', '')
        # print('...check annotations for %s' % tax)
        jf = '%s/%s.json' % (annotation_dir, tax)
        gf = '%s/%s/%s.fa' % (searchTaxa_dir, tax, tax)
        jobs.append([gf,jf, annotation_dir])
    cpus = mp.cpu_count()-1
    pool = mp.Pool(cpus)
    out = []
    for _ in tqdm(pool.imap_unordered(check_complete_anno, jobs), total=len(jobs)):
        out.append(_)
    return None


def check_missing_ncbiID(taxon_list):
    """ Check all taxa in searchTaxa_dir and coreTaxa_dir
    if they are have valid NCBI taxonomy IDs
    """
    ncbi = NCBITaxa()
    missing_taxa = {}
    present_taxa = {}
    dup_taxa = []
    for t in taxon_list:
        tax_id = t.split('@')[1]
        try:
            taxid2name = ncbi.get_taxid_translator([tax_id])
            if len(taxid2name) < 1:
                if not t+'\t'+str(tax_id) in missing_taxa:
                    missing_taxa[t+'\t'+str(tax_id)] = 1
        except:
            if not t+'\t'+str(tax_id) in missing_taxa:
                missing_taxa[t+'\t'+str(tax_id)] = 1
        if not tax_id in present_taxa:
            present_taxa[tax_id] = t
        else:
            dup_taxa.append('%s\t%s' % (t, present_taxa[tax_id]))
    return(missing_taxa.keys(), dup_taxa)


def run_check_missing_ncbiID():
    """ Run check_missing_ncbiID fn """
    pass


def main():
    version = get_distribution('fdog').version
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(version) + '.')
    parser.add_argument('-s', '--searchTaxa_dir', help='Path to search taxa directory (e.g. fdog_dataPath/searchTaxa_dir)', action='store', default='')
    parser.add_argument('-c', '--coreTaxa_dir', help='Path to blastDB directory (e.g. fdog_dataPath/coreTaxa_dir)', action='store', default='')
    parser.add_argument('-a', '--annotation_dir', help='Path to feature annotation directory (e.g. fdog_dataPath/annotation_dir)', action='store', default='')
    parser.add_argument('--replace', help='Replace special characters in sequences by "X"', action='store_true', default=False)
    parser.add_argument('--delete', help='Delete special characters in sequences', action='store_true', default=False)
    parser.add_argument('--concat', help='Concatenate multiple-line sequences into single-line', action='store_true', default=False)
    parser.add_argument('--reblast', help='Re-create blast databases', action='store_true', default=False)

    ### get arguments
    args = parser.parse_args()

    searchTaxa_dir = args.searchTaxa_dir
    coreTaxa_dir = args.coreTaxa_dir
    annotation_dir = args.annotation_dir
    replace = args.replace
    delete = args.delete
    concat = args.concat
    reblast = args.reblast

    checkOptConflict(concat, replace, delete)
    caution = 0

    ### get fdog dir and assign searchTaxa_dir, coreTaxa_dir, annotation_dir if not given
    fdogPath = os.path.realpath(__file__).replace('/checkData.py','')
    if not searchTaxa_dir or not coreTaxa_dir or not annotation_dir:
        pathconfigFile = fdogPath + '/bin/pathconfig.txt'
        if not os.path.exists(pathconfigFile):
            sys.exit('No pathconfig.txt found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
        with open(pathconfigFile) as f:
            dataPath = f.readline().strip()
        if not searchTaxa_dir:
            searchTaxa_dir = dataPath + "/searchTaxa_dir"
        if not coreTaxa_dir:
            coreTaxa_dir = dataPath + "/coreTaxa_dir"
        if not annotation_dir:
            annotation_dir = dataPath + "/annotation_dir"

    searchTaxa_dir = os.path.abspath(searchTaxa_dir)
    coreTaxa_dir = os.path.abspath(coreTaxa_dir)
    annotation_dir = os.path.abspath(annotation_dir)

    ### check searchTaxa_dir
    print('=> Checking %s...' % searchTaxa_dir)
    genomeTaxa = run_check_fasta(searchTaxa_dir, replace, delete, concat)

    ### check coreTaxa_dir
    if reblast:
        print('=> (Re-)Creating blastDBs...')
        failed_blast = run_create_blastdb(coreTaxa_dir, searchTaxa_dir)
        if len(failed_blast) > 0:
            print('*** WARNING: Some BlastDBs cannot be created:\n%s' % ', '.join(failed_blast))
        else:
            print('All old BlastDBs have been updated!')
    print('=> Checking %s...' % coreTaxa_dir)
    blastTaxa = run_check_fasta(coreTaxa_dir, replace, delete, concat)
    check_blast = run_check_blastdb(coreTaxa_dir, fdogPath)

    if not check_blast[0] == 1:
        print('*** ERROR: Version incompatible between BlastDB and BLAST program!')
        print('For checking, run: blastp -query %s -db %s' % (check_blast[0], check_blast[1]))
        print('Consider using --reblast option to update old BlastDBs!')
        sys.exit()

    ### check annotation_dir
    print('=> Checking %s...' % annotation_dir)
    missing_anno = check_missing_json(annotation_dir, general_fn.join_2lists(genomeTaxa, blastTaxa))
    if len(missing_anno) > 0:
        print('\033[92m*** WARNING: Annotation files not found for:\033[0m')
        print(*missing_anno, sep = "\n")
        print('NOTE: You still can run fdog without FAS using the option "-fasoff"')
        caution = 1
    run_check_complete_anno(annotation_dir, searchTaxa_dir)

    # ### check ncbi IDs
    print('=> Checking NCBI taxonomy IDs...')
    missing_taxa, dup_taxa = check_missing_ncbiID(general_fn.join_2lists(genomeTaxa, blastTaxa))
    if (len(missing_taxa) > 0):
        print('\033[92m*** WARNING: Taxa not found in current local NCBI taxonomy database:\033[0m')
        print(*missing_taxa, sep = "\n")
        print('==> NOTE: You still can run fDOG with those taxa, but they will not be included in the core set compilation!')
        caution = 1
    if (len(dup_taxa) > 0):
        print('\033[92m*** WARNING: These taxa have the same NCBI taxonomy IDs:\033[0m')
        print(*dup_taxa, sep = "\n")
        print('==> NOTE: This could lead to some conflicts!')
        caution = 1

    print('---------------------------------')
    if caution == 1:
        print('==> Done! Data are ready to use WITH CAUTION!')
    else:
        print('==> Done! Data are ready to use!')

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from pathlib import Path
from Bio import SeqIO

import fdog.libs.zzz as general_fn
import fdog.libs.fasta as fasta_fn
import fdog.libs.blast as blast_fn
import fdog.libs.output as output_fn


##### FUNCTIONS FOR DATA/INPUT PREPARATION #####

def parsing_paths(args):
    """ Getting path to hmm core set, coreTaxa_dir, searchTaxa_dir and annotation_dir"""
    (pathFile, outpath, hmmpath, corepath, searchpath, annopath) = args
    ### get fdog and data path
    data_path = ''
    fdog_path = os.path.realpath(__file__).replace('/libs/preparation.py','')
    pathconfig_file = fdog_path + '/bin/pathconfig.txt'
    # pathconfig_file = '/home/vinh/anaconda3/envs/test_fas/lib/python3.9/site-packages/fdog/bin/pathconfig.txt' ######################################  REMOVE THIS
    if not os.path.exists(pathconfig_file):
        sys.exit(
            'No pathconfig.txt found at %s. Please run fdog.setup '
            + '(https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).'
            % pathconfig_file)
    if pathFile == '':
        with open(pathconfig_file) as f:
            data_path = f.readline().strip()
    else:
        cfg = general_fn.load_config(pathFile)
        try:
            data_path = cfg['dataPath']
        except:
            data_path = 'config'

    if hmmpath == '':
        hmmpath = outpath + '/core_orthologs'
        Path(hmmpath).mkdir(parents = True, exist_ok = True)

    if corepath == '':
        corepath = data_path + '/coreTaxa_dir'
        if data_path == 'config':
            try:
                corepath = cfg['corepath']
            except:
                sys.exit('corepath not found in %s' % pathFile)
    if searchpath == '':
        searchpath = data_path + '/searchTaxa_dir'
        if data_path == 'config':
            try:
                searchpath = cfg['searchpath']
            except:
                sys.exit('searchpath not found in %s' % pathFile)
    if annopath == '':
        annopath = data_path + '/annotation_dir'
        if data_path == 'config':
            try:
                annopath = cfg['annopath']
            except:
                sys.exit('annopath not found in %s' % pathFile)
    return(hmmpath, corepath, searchpath, annopath)


def check_input(args):
    (seqFile, refspec, outpath, hmmpath, corepath,
        searchpath, annopath, pathFile) = args
    fdog_path = os.path.realpath(__file__).replace('/libs/preparation.py','')
    # create output directory
    Path(outpath).mkdir(parents = True, exist_ok = True)
    Path(hmmpath).mkdir(parents = True, exist_ok = True)
    # check path existing
    hmmpath, corepath, searchpath, annopath = parsing_paths(
        [pathFile, outpath, hmmpath, corepath, searchpath, annopath])
    for path in [hmmpath, corepath, searchpath, annopath]:
        general_fn.check_file_exist(path)
    # check for seqFile
    if not os.path.exists(os.path.abspath(seqFile)):
        if not os.path.exists(fdog_path + '/data/' + seqFile):
            sys.exit(
                '%s not found in %s or %s'
                % (seqFile, os.getcwd(), fdog_path + '/data/'))
        else:
            seqFile = fdog_path + '/data/' + seqFile
    else:
        seqFile = os.path.abspath(seqFile)
    # check refspec
    if not os.path.exists(os.path.abspath(corepath+'/'+refspec)):
        exit('Reference taxon %s not found in %s' % (refspec, corepath))
    return (seqFile, hmmpath, corepath, searchpath, annopath)


def identify_seed_id(seqFile, refspec, corepath, debug, silentOff):
    refspec_db = '%s/%s/%s' % (corepath, refspec, refspec)
    # first check if input seed ID existiert in refspec genome
    refspec_fa = fasta_fn.read_fasta('%s.fa' % refspec_db)
    seed_fa = SeqIO.parse(open(seqFile),'fasta')
    for seed in seed_fa:
        try:
            if len(refspec_fa.fetch(seed.id)) == len(seed.seq):
                return(seed.id)
        except:
            output_fn.print_debug(debug, 'Identify seed ID', 'Input seed ID not found!')
    # otherwise, perform blast search
    blast_xml = blast_fn.do_blastsearch(seqFile, refspec_db)
    blast_out = blast_fn.parse_blast_xml(blast_xml)
    for hit in blast_out['hits']:
        if blast_out['hits'][hit]['align_len'] == blast_out['query_len']:
            return(hit)
        elif abs(int(blast_out['hits'][hit]['align_len']) - int(blast_out['query_len'])) < 10:
            output_fn.print_stdout(silentOff, 'WARNING: Found seed sequence shorter than input!')
            return(hit)
        else:
            sys.exit('ERROR: Cannot find seed sequence in genome of reference species for %s!' % blast_out['query'])



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import re
from ete3 import NCBITaxa

import fdog.libs.zzz as general_fn


##### FUNCTIONS RELATED TO TAXONOMY TREE #####

def get_rank_index(lineage, rank_name, ncbi):
    """ Get ID and index in the species lineage for a given rank
    Return {rank_id:rank_index}
    """
    ranks = ncbi.get_rank(lineage)
    rank_id = list(general_fn.matching_elements(ranks, rank_name).keys())[0]
    rank_index = len(ranks) - lineage.index(rank_id) - 1
    return({rank_id:rank_index})


def get_rank_range(lineage, minDist, maxDist, ncbi):
    """ Get rank ID and its index in a given species lineage
    for a pair of min and max rank. See get_rank_index()
    Return a list of 2 dictionary for min and max rank as
    [{min_rank_id:min_rank_index}, {max_rank_id:max_rank_index}]
    """
    return(
        get_rank_index(lineage, minDist, ncbi),
        get_rank_index(lineage, maxDist, ncbi))


def check_taxon_group(group_id, tax_id, ncbi):
    """ Check if a taxon (tax_id) belongs to a taxonomy group (group_id)"""
    lineage = ncbi.get_lineage(tax_id)
    if group_id in lineage:
        return(True)
    return(False)


def get_ancestor(id1, id2, ncbi):
    """ Get common ancestor ID and rank for 2 taxon IDs
    Return dictionary {ancestor_id: ancestor_rank}
    """
    tree = ncbi.get_topology([id1, id2], intermediate_nodes = False)
    ancestor = tree.get_common_ancestor(id1, id2).name
    return(ncbi.get_rank([ancestor]))


def check_common_ancestor(ref_id, ancestor, minDist, maxDist, ncbi):
    """ Check if ancestor ID lies within the range between min and max rank
    of reference species
    Return 1 if true
    """
    ref_lineage = ncbi.get_lineage(ref_id)
    (min_ref, max_ref) = get_rank_range(ref_lineage, minDist, maxDist, ncbi)
    ancestor_index = len(ref_lineage) - ref_lineage.index(ancestor) - 1
    if list(min_ref.values())[0] <= ancestor_index <= list(max_ref.values())[0]:
        return(1)
    return(0)


def remove_clade(tree, node_id):
    """ Remove a clade from a tree """
    removed_clade = tree.search_nodes(name = str(node_id))[0]
    removed_node = removed_clade.detach()
    return(tree)


def get_leaves_dict(spec_lineage, tree, min_index, max_index):
    """ Given a tree and a lineage string of a species
    Return a dictionary where keys are the internal nodes defined by the
    ranks between min rank (e.g. genus, specified by min_index in the species
    lineage) and max rank (e.g. phylum). Values are all leaves in the tree
    that belong to the corresponding internal node (rank)
    """
    node_dict = {}
    already_added = []
    spec_lineage.reverse()
    for i in range(len(spec_lineage)):
        if i >= min_index and i <= max_index:
            curr_node = spec_lineage[i]
            node = tree.search_nodes(name = str(curr_node))
            if len(node) > 0:
                for leaf in node:
                    node_dict[spec_lineage[i]] = []
                    for t in leaf.traverse():
                        if t.is_leaf():
                            if not t.name in already_added:
                                already_added.append(t.name)
                                node_dict[spec_lineage[i]].append(t.name)
    return(general_fn.remove_dup_in_dict(node_dict))


def check_tax_id(tax_id):
    """ Check valid taxon ID
    Return taxon name (UNK<tax_id> if ID not found in ncbi db)
    """
    ncbi = NCBITaxa()
    tmp = ncbi.get_rank([tax_id])
    try:
        tmp = ncbi.get_rank([tax_id])
        rank = tmp[int(tax_id)]
        if not rank == 'species':
            print('\033[92mWARNING: rank of %s is not SPECIES (%s)\033[0m' % (tax_id, rank))
        else:
            ncbi_name = ncbi.get_taxid_translator([tax_id])[int(tax_id)]
            print('\033[92mNCBI taxon info: %s %s\033[0m' % (tax_id, ncbi_name))
        return(ncbi_name)
    except:
        print('\033[92mWARNING: %s not found in NCBI taxonomy database!\033[0m' % tax_id)
        return('UNK%s' % tax_id)


def abbr_ncbi_name(ncbi_name):
    """ Parse ncbi taxon name into abbr name
    E.g. "Homo sapiens" -> "HOMSA"
    """
    if not ncbi_name.startswith('UNK'):
        ncbi_name = re.sub('[^a-zA-Z1-9\s]+', '', ncbi_name)
        tax_name = ncbi_name.split()
        name = tax_name[0][:3].upper()+tax_name[1][:2].upper()
    else:
        name = ncbi_name
    return(name)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to prepare data for fdog.
#  For each given genome FASTA file, It will create a folder within searchTaxa_dir
#  with the naming scheme of fdog ([Species acronym]@[NCBI ID]@[Proteome version]
#  e.g HUMAN@9606@3), a annotation file in JSON format in annotation_dir and
#  a blast DB in coreTaxa_dir folder (optional).
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import argparse
from pathlib import Path
import shutil
import multiprocessing as mp
from datetime import datetime
from pkg_resources import get_distribution

import fdog.libs.zzz as general_fn
import fdog.libs.tree as tree_fn
import fdog.libs.addtaxon as add_taxon_fn


def main():
    version = get_distribution('fdog').version
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(version) + '.')
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('-f', '--fasta', help='FASTA file of input taxon', action='store', default='', required=True)
    required.add_argument('-i', '--taxid', help='Taxonomy ID of input taxon', action='store', default='', required=True, type=int)
    optional.add_argument('-o', '--outPath', help='Path to output directory', action='store', default='')
    optional.add_argument('-n', '--name', help='Acronym name of input taxon', action='store', default='', type=str)
    optional.add_argument('-v', '--verProt', help='Proteome version', action='store', default='', type=str)
    optional.add_argument('-c', '--coreTaxa', help='Include this taxon to core taxa (i.e. taxa in coreTaxa_dir folder)', action='store_true', default=False)
    optional.add_argument('-a', '--noAnno', help='Do NOT annotate this taxon using fas.doAnno', action='store_true', default=False)
    optional.add_argument('--cpus', help='Number of CPUs used for annotation. Default = available cores - 1', action='store', default=0, type=int)
    optional.add_argument('--replace', help='Replace special characters in sequences by "X"', action='store_true', default=False)
    optional.add_argument('--delete', help='Delete special characters in sequences', action='store_true', default=False)
    optional.add_argument('--force', help='Force overwrite existing data', action='store_true', default=False)

    args = parser.parse_args()

    general_fn.check_file_exist(args.fasta)
    faIn = args.fasta
    taxId = str(args.taxid)
    outPath = args.outPath
    if outPath == '':
        fdogPath = os.path.realpath(__file__).replace('/addTaxon.py','')
        pathconfigFile = fdogPath + '/bin/pathconfig.txt'
        if not os.path.exists(pathconfigFile):
            sys.exit('No pathconfig.txt found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
        with open(pathconfigFile) as f:
            outPath = f.readline().strip()
    outPath = os.path.abspath(outPath)
    name = args.name.upper()
    ver = str(args.verProt)
    if ver == '':
        ver = datetime.today().strftime('%y%m%d')
    noAnno = args.noAnno
    coreTaxa = args.coreTaxa
    cpus = args.cpus
    if cpus == 0:
        cpus = mp.cpu_count()-2
    replace = args.replace
    delete = args.delete
    add_taxon_fn.check_conflict_opts(replace, delete)
    force = args.force

    ### species name after fdog naming scheme
    spec_name = add_taxon_fn.generate_spec_name(taxId, name, ver)
    print('Species name\t%s' % spec_name)

    ### remove old folder if force is set
    if force == True:
        if os.path.exists(outPath + '/searchTaxa_dir/' + spec_name):
            shutil.rmtree(outPath + '/searchTaxa_dir/' + spec_name)
        if os.path.exists(outPath + '/coreTaxa_dir/' + spec_name):
            shutil.rmtree(outPath + '/coreTaxa_dir/' + spec_name)

    ### initiate paths
    genome_path = add_taxon_fn.create_folders(outPath, spec_name, coreTaxa, noAnno)

    ### create file in searchTaxa_dir
    print('Parsing FASTA file...')
    genome_file = add_taxon_fn.create_genome([faIn, genome_path, spec_name, force, replace, delete])
    out_msg = 'Output for %s can be found in %s within searchTaxa_dir'  % (spec_name, outPath)

    ### create blast db
    if coreTaxa:
        print('\nCreating Blast DB...')
        add_taxon_fn.create_blastdb([outPath, spec_name, genome_file, force, False])
        out_msg = '%s, coreTaxa_dir' % out_msg

    ### create annotation
    if not noAnno:
        add_taxon_fn.create_annoFile(outPath, genome_file, cpus, force)
        if os.path.exists('%s/annotation_dir/tmp' % outPath):
            shutil.rmtree('%s/annotation_dir/tmp' % outPath)
        out_msg = '%s, annotation_dir' % out_msg

    print('\n==> %s' % out_msg)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to uninstall fdog and its data
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import argparse
import subprocess
import shutil
from pkg_resources import get_distribution


def query_yes_no(question, default='yes'):
    valid = {'yes': True, 'y': True, 'ye': True,
             'no': False, 'n': False}
    if default is None:
        prompt = ' [y/n] '
    elif default == 'yes':
        prompt = ' [Y/n] '
    elif default == 'no':
        prompt = ' [y/N] '
    else:
        raise ValueError('invalid default answer: "%s"' % default)
    while True:
        # sys.stdout.write(question + prompt)
        choice = sys.stdin.readline().rstrip().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write('Please respond with "yes" or "no" '
                             '(or "y" or "n").\n')


def main():
    version = get_distribution('fdog').version
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(version) + '.')
    parser.add_argument('--data', help='Remove fdog together with all files/data within the installed fdog directory', action='store_true', default=False)
    args = parser.parse_args()
    data = args.data

    fdogPath = os.path.realpath(__file__).replace('/removefDog.py','')
    pathconfigFile = fdogPath + '/bin/pathconfig.txt'
    if not os.path.exists(pathconfigFile):
        sys.exit('No pathconfig.txt found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
    with open(pathconfigFile) as f:
        dataPath = f.readline().strip()

    if data:
        print('All files and folders in %s will be removed! Enter to continue' % fdogPath)
    else:
        print('fdog will be uninstalled. Some files/data still can be found in %s! Enter to continue' % fdogPath)
    if query_yes_no('Are you sure?'):
        if data:
            folders = ['bin', 'data']
            for f in folders:
                dirPath = fdogPath+'/'+f
                if os.path.exists(os.path.abspath(dirPath)):
                    print('removing %s...' % f)
                    shutil.rmtree(dirPath)
        uninstallCmd = 'pip uninstall fdog'
        try:
            subprocess.call([uninstallCmd], shell = True)
        except:
            print('Error by uninstalling fdog. Please manually uninstall it using <pip uninstall fdog>')
        if data:
            if os.path.exists(os.path.abspath(fdogPath)):
                shutil.rmtree(fdogPath)

    print('NOTE: fdog genome data are still available at %s.' % dataPath)


if __name__ == '__main__':
    main()
